/*
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this code; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  */

/* Copyright (C) 2002 Ian Zimmerman <itz@speakeasy.org> */

static const char rcsid[] = "$Id: my_clist.cc,v 1.1 2002/09/09 18:30:07 itz Exp $ ";

#include "my_clist.hh"

#include <fstream>
#include <cmath>
#include <algorithm>
#include <functional>
#include <cstdio>

struct sys_color {
  int index;
  double red;
  double green;
  double blue;
  string name;
  double brightness;
  double distance;
};

struct distance_calc_fun {
  double red;
  double green;
  double blue;
  distance_calc_fun(double red, double green, double blue);
  void apply(sys_color* p_color);
  void operator() (sys_color* p_color);
};  

distance_calc_fun::distance_calc_fun(double red, double green, double blue)
  : red(red), green(green), blue(blue)
{}

void
distance_calc_fun::apply(sys_color* p_color)
{
  double d_red = red - p_color->red;
  double d_green = green - p_color->green;
  double d_blue = blue - p_color->blue;
  p_color->distance = sqrt
    (d_red * d_red + d_green * d_green + d_blue * d_blue);
}

void
distance_calc_fun::operator() (sys_color* p_color)
{
  apply(p_color);
}

struct does_occur_fun {
  const char* s;
  does_occur_fun(const char* s);
  int apply(sys_color* p_color);
  int operator() (sys_color* p_color);
};

does_occur_fun::does_occur_fun(const char* s)
  : s(s)
{}

int
does_occur_fun::apply(sys_color* p_color)
{
  return (p_color->name.find(s) != string::npos);
}

int 
does_occur_fun::operator() (sys_color* p_color)
{
  return apply(p_color);
}

static int
num_less(sys_color* p1, sys_color* p2)
{
  return (p1->index < p2->index);
}

static int
num_greater(sys_color* p1, sys_color* p2)
{
  return (p1->index > p2->index);
}

static int
name_less(sys_color* p1, sys_color* p2)
{
  return (p1->name < p2->name);
}

static int
name_greater(sys_color* p1, sys_color* p2)
{
  return (p1->name > p2->name);
}

static int
bright_less(sys_color* p1, sys_color* p2)
{
  return (p1->brightness < p2->brightness);
}

static int
bright_greater(sys_color* p1, sys_color* p2)
{
  return (p1->brightness > p2->brightness);
}

static int
dist_less(sys_color* p1, sys_color* p2)
{
  return (p1->distance < p2->distance);
}

static int
dist_greater(sys_color* p1, sys_color* p2)
{
  return (p1->distance > p2->distance);
}

static const char* titles[] = { "#", "System color", "Brightness", "Distance", 0 };

my_clist::my_clist(const char* rgb_file_name)
  : Gtk::CList(4, titles), has_selection(0), sorted_by(num_inc), last_found(-1)
{
  set_column_auto_resize(0, 1);
  set_column_auto_resize(1, 1);

  // initialize vector of system colors
  colors.reserve(1000);
  
  // fill vector with known system color names
  int index = 0;
  ifstream rgb_istr(rgb_file_name);
  if (!rgb_istr) {
    cerr << "Cannot open " << rgb_file_name << endl;
  } else {
    string comment;
    int c = rgb_istr.peek();
    while ((rgb_istr.rdstate() & ios::eofbit) == 0) {
      if (c == '!') {
        getline(rgb_istr, comment);
        c = rgb_istr.peek();
        continue;
      } /*if*/
      int red;
      int green;
      int blue;
      string rest;
      rgb_istr >> red >> green >> blue;
      sys_color* new_color = new sys_color;
      new_color->red = (double)red/(double)255;
      new_color->green = (double)green/(double)255;
      new_color->blue = (double)blue/(double)255;
      new_color->brightness = sqrt
        ((new_color->red * new_color->red
          + new_color->green * new_color->green
          + new_color->blue * new_color->blue) / 3) ;
      getline(rgb_istr, rest);
      int real = rest.find_first_not_of(" \t");
      new_color->name = rest.substr(real);
      new_color->index = index++;
      colors.insert(colors.end(), new_color);
      c = rgb_istr.peek();
    } /*while*/
  } /*if*/
}

void
my_clist::fill_internal()
{

  freeze();
  clear();
  has_selection = 0;
  for (vector<sys_color*>::iterator iter = colors.begin(); iter != colors.end(); ++iter) {
    char index_buf[10];
    char brightness_buf[10];
    char distance_buf[10];
    snprintf(index_buf, 10, "%3d", (*iter)->index+1);
    snprintf(brightness_buf, 10, "%f", (*iter)->brightness);
    snprintf(distance_buf, 10, "%f", (*iter)->distance);
  
    static const char* cells[5];
    cells[0] = index_buf;
    cells[1] = (*iter)->name.c_str();
    cells[2] = brightness_buf;
    cells[3] = distance_buf;
    cells[4] = 0;
  
    append_row(cells);
  } /*for*/
  thaw();

}

void
my_clist::fill(gdouble red, gdouble green, gdouble blue)
{
  for_each(colors.begin(), colors.end(), distance_calc_fun(red, green, blue));
  fill_internal();
}

void
my_clist::color_change(gdouble red, gdouble green, gdouble blue)
{
  for_each(colors.begin(), colors.end(), distance_calc_fun(red, green, blue));
  freeze();
  int i = 0;
  for (vector<sys_color*>::iterator iter = colors.begin();
       iter != colors.end(); ++iter, ++i) {
    char distance_buf[10];
    snprintf(distance_buf, 10, "%f", (*iter)->distance);
    set_text(i, 3, distance_buf);
  } /*for*/
  
  thaw();
  if (sorted_by == dist_inc || sorted_by == dist_dec) {
    sorted_by = none;
  } /*if*/
}

void
my_clist::select_row_impl(gint row, gint column, GdkEvent* p)
{
  inherited::select_row_impl(row, column, p);

  has_selection = 1;
  selected_color[0] = colors[row]->red;
  selected_color[1] = colors[row]->green;
  selected_color[2] = colors[row]->blue;
  selected_color[3] = 0.0;
}

void
my_clist::unselect_row_impl(gint row, gint column, GdkEvent* p)
{
  has_selection = 0;
  inherited::unselect_row_impl(row, column, p);
}

const gdouble*
my_clist::get_selected_color()
{
  if (!has_selection) {
    return 0;
  } else {
    return selected_color;
  } /*if*/
}

void
my_clist::sort_by_num_inc()
{
  ::sort(colors.begin(), colors.end(), num_less);
  fill_internal();
  sorted_by = num_inc;
}

void
my_clist::sort_by_num_dec()
{
  ::sort(colors.begin(), colors.end(), num_greater);
  fill_internal();
  sorted_by = num_dec;
}

void
my_clist::sort_by_name_inc()
{
  ::sort(colors.begin(), colors.end(), name_less);
  fill_internal();
  sorted_by = name_inc;
}

void
my_clist::sort_by_name_dec()
{
  ::sort(colors.begin(), colors.end(), name_greater);
  fill_internal();
  sorted_by = name_dec;
}

void
my_clist::sort_by_bright_inc()
{
  ::sort(colors.begin(), colors.end(), bright_less);
  fill_internal();
  sorted_by = bright_inc;
}
 
void
my_clist::sort_by_bright_dec()
{
  ::sort(colors.begin(), colors.end(), bright_greater);
  fill_internal();
  sorted_by = bright_dec;
}

void
my_clist::sort_by_dist_inc()
{
  ::sort(colors.begin(), colors.end(), dist_less);
  fill_internal();
  sorted_by = dist_inc;
}

void 
my_clist::sort_by_dist_dec()
{
  ::sort(colors.begin(), colors.end(), dist_greater);
  fill_internal();
  sorted_by = dist_dec;
}

void
my_clist::click_column_impl(gint column)
{
  switch(column) {
  case 0:
    if (sorted_by == num_inc) sort_by_num_dec();
    else sort_by_num_inc();
    break;
  case 1:
    if (sorted_by == name_inc) sort_by_name_dec();
    else sort_by_name_inc();
    break;
  case 2:
    if (sorted_by == bright_inc) sort_by_bright_dec();
    else sort_by_bright_inc();
    break;
  case 3:
    if (sorted_by == dist_inc) sort_by_dist_dec();
    else sort_by_dist_inc();
    break;
  } /*switch*/
}

int
my_clist::find_first(const char* s)
{
  last_found = -1;
  return find_next(s);
}

int
my_clist::find_next(const char* s)
{
  vector<sys_color*>::iterator iter = colors.begin() + (last_found+1);
  does_occur_fun pred(s);
  vector<sys_color*>::iterator where = ::find_if(iter, colors.end(), pred);
  if (where != colors.end()) {
    int row_idx = where - colors.begin();
    moveto(row_idx, 0);
    row(row_idx).select();
    last_found = row_idx;
    return 1;
  } else {
    return 0;
  }
}
